'''
exp_conv_irf:
submodule for the mathematical functions for 
exponential decay convolved with irf

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3.
'''

import numpy as np
from scipy.special import erfc, erfcx, exp1


def exp_conv_gau(t, fwhm, k):

    '''
    Compute exponential function convolved with normalized gaussian 
    distribution

    Note.
    We assume temporal pulse of x-ray is normalized gaussian distribution
    
    .. math::

       \sigma = \\frac{fwhm}{2\sqrt{2\log{2}}}


    .. math::

       IRF(t) = \\frac{1}{\\sigma \\sqrt{2\pi}}\\exp\\left(-\\frac{t^2}{2\\sigma^2}\\right)

    :param t: time
    :type t: float, numpy_1d_array
    :param fwhm: full width at half maximum of x-ray temporal pulse
    :type fwhm: float
    :param k: rate constant (inverse of life time)
    :type k: float

    
    :return: convolution of normalized gaussian distribution and exp(-kt)

    .. math::

       \\frac{1}{2}\exp\\left(\\frac{k^2}{2\sigma^2}-kt\\right){erfc}\\left(\\frac{1}{\sqrt{2}}\\left(k\sigma-\\frac{t}{\sigma}\\right)\\right)


    :rtype: numpy_1d_array
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    ksigma = k*sigma
    ksigma2 = ksigma*sigma
    tinvsigma = t/sigma

    if not isinstance(t, np.ndarray):
        if ksigma2 > t:
            ans = 1/2*np.exp(-0.5*tinvsigma**2) * \
                erfcx(1/np.sqrt(2)*(ksigma-tinvsigma))
        else:
            ans = 1/2*np.exp(k*(0.5*ksigma2-t)) * \
                erfc(1/np.sqrt(2)*(ksigma - tinvsigma))
    else:
        mask = t < ksigma2
        inv_mask = np.invert(mask)
        ans = np.zeros(t.shape[0])
        ans[mask] = 1/2*np.exp(-0.5*tinvsigma[mask]**2) * \
            erfcx(1/np.sqrt(2)*(ksigma-tinvsigma[mask]))
        ans[inv_mask] = 1/2*np.exp(k*(0.5*ksigma2-t[inv_mask])) * \
            erfc(1/np.sqrt(2)*(ksigma - tinvsigma[inv_mask]))

    return ans


def exp_conv_cauchy(t, fwhm, k):

    '''
    Compute exponential function convolved with normalized cauchy 
    distribution

    Note.
    We assume temporal pulse of x-ray is normalized cauchy distribution

    .. math::
       \\gamma = \\frac{fwhm}{2}

    .. math::
       IRF(t) = \\frac{\\gamma}{\\pi}\\frac{1}{(x-t)^2+\\gamma^2}
    

    :param t: time
    :type t: float, numpy_1d_array
    :param fwhm: full width at half maximum of x-ray temporal pulse
    :type fwhm: float
    :param k: rate constant (inverse of life time)
    :type k: float

   
    :return: convolution of normalized cauchy distribution and exp(-kt)

    .. math::
       \\frac{\\exp(-kt)}{\\pi}\\Im\\left(\\exp(-ik\\gamma)\cdot{E1}(-kt-ik\\gamma)\\right)


    :rtype: numpy_1d_array 
    '''

    gamma = fwhm/2
    if k == 0:
        ans = 0.5+1/np.pi*np.arctan(t/gamma)
    else:
        ikgamma = complex(0, k*gamma)
        kt = k*t
        if not isinstance(t, np.ndarray):
            if np.abs(kt) < 700:
                ans = np.exp(-ikgamma)*exp1(-kt-ikgamma)
                ans = np.exp(-kt)*ans.imag/np.pi
            else:
                inv_z = 1/(kt+ikgamma)
                ans = 1+10*inv_z
                ans = 1+9*inv_z*ans
                ans = 1+8*inv_z*ans
                ans = 1+7*inv_z*ans
                ans = 1+6*inv_z*ans
                ans = 1+5*inv_z*ans
                ans = 1+4*inv_z*ans
                ans = 1+3*inv_z*ans
                ans = 1+2*inv_z*ans
                ans = 1+inv_z*ans
                ans = -inv_z*ans
                ans = ans.imag/np.pi
        else:
            mask = np.abs(kt) < 700
            inv_mask = np.invert(mask)
         
            ans = np.zeros(t.shape[0], dtype='float')
            inv_z = 1/(kt[inv_mask]+ikgamma)

            # abs(kt) < 700
            ans1 = np.exp(-ikgamma)*exp1(-kt[mask]-ikgamma)
            ans[mask] = np.exp(-kt[mask])*ans1.imag/np.pi

            # abs(kt) > 700, use asymptotic series
            ans2 = 1+10*inv_z
            ans2 = 1+9*inv_z*ans2
            ans2 = 1+8*inv_z*ans2
            ans2 = 1+7*inv_z*ans2
            ans2 = 1+6*inv_z*ans2
            ans2 = 1+5*inv_z*ans2
            ans2 = 1+4*inv_z*ans2
            ans2 = 1+3*inv_z*ans2
            ans2 = 1+2*inv_z*ans2
            ans2 = 1+1*inv_z*ans2
            ans2 = -inv_z*ans2
            ans[inv_mask] = ans2.imag/np.pi
    return ans


def exp_conv_pvoigt(t, fwhm_G, fwhm_L, eta, k):

    '''
    Compute exponential function convolved with normalized pseudo
    voigt profile

    Note.
    We assume temporal pulse of x-ray is pseudo voigt profile

    .. math::
       \mathrm{pvoigt}(t) = (1-\\eta)G({fwhm}_G, t)+\\eta C({fwhm}_L, t)
    

    :param t: time
    :type t: float, numpy_1d_array
    :param fwhm_G: full width at half maximum of x-ray temporal pulse gaussian part
    :type fwhm_G: float
    :param fwhm_L: full width at half maximum of x-ray temporal pulse cauchy part
    :type fwhm_L: float
    :param eta: mixing parameter
    :type eta: float
    :param k: rate constant (inverse of life time)
    :type k: float

   
    :return: convolution of normalized pseudo voigt profile and exp(-kt)
    :rtype: numpy_1d_array 
    '''

    return eta*exp_conv_cauchy(t, fwhm_L, k) + \
        (1-eta)*exp_conv_gau(t, fwhm_G, k)



'''
data_process:

subpackage for data processing

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3.
'''

from .scale import automate_scaling
from .corr import corr_a_method

__all__ = ['automate_scaling', 'corr_a_method']



'''
A_matrix:
submodule for evaluation of A_matrix

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3.
'''
import numpy as np
from .exp_conv_irf import exp_conv_gau, exp_conv_cauchy


def make_A_matrix(t, k):

    A = np.zeros((k.shape[0], t.shape[0]))
    for i in range(k.shape[0]):
        A[i, :] = np.exp(-k[i]*t)
    A = A*np.heaviside(t, 1)
    A[np.isnan(A)] = 0
    
    return A


def make_A_matrix_gau(t, fwhm, k):

    A = np.zeros((k.shape[0], t.shape[0]))
    for i in range(k.shape[0]):
        A[i, :] = exp_conv_gau(t, fwhm, k[i])
    
    return A


def make_A_matrix_cauchy(t, fwhm, k):

    A = np.zeros((k.shape[0], t.shape[0]))
    for i in range(k.shape[0]):
        A[i, :] = exp_conv_cauchy(t, fwhm, k[i])

    return A


def make_A_matrix_pvoigt(t, fwhm_G, fwhm_L, eta, k):
    return (1-eta)*make_A_matrix_gau(t, fwhm_G, k) + \
        eta*make_A_matrix_cauchy(t, fwhm_L, k)




'''
corr:

submodule for correction of scale using one time delay scan

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3.
'''
import numpy as np
from scipy.interpolate import interp1d


def corr_a_method(e_ref_index, e, t,
                  escan_time, ref_tscan_energy, ref_time_zeros,
                  escan_data=None, escan_data_eps=None,
                  ref_tscan_data=None, ref_tscan_data_eps=None,
                  warn=False):

    '''
    corr_a_method:
    Corrects the scaling of escan using one time delay scan 

    :param int e_ref: index of reference escab used for ``A-method``
    :param numpy_1d_array e: array of energies in which we measured escan
    :param numpy_1d_array t: array of time delays in which we measured tscan
    :param numpy_1d_array escan_time: 
     array of time delays at which we measured escan
    :param float ref_tscan_energy: 
     reference energy for repairing scale of escan
    :param float ref_time_zeros:
     time zero for reference tscan
    :param numpy_nd_array escan_data: data for escan
     (Note. escan data does not contains energy range)
    :param numpy_nd_array escan_data_eps: error for escan data 
    :param numpy_1d_array tscan_data: data for reference tscan
    :param numpy_1d_array tscan_data_eps: error for reference tscan 

    :return: corrected_data
    :rtype: dict
    :return: corrected_data['escan'] : corrected data for escan
    :rtype: numpy_nd_array
    :return: corrected_data['escan_eps'] : corrected error for escan
    :rtype: numpy_nd_array
    '''

    # init scaled_data
    scaled_data = dict()

    # scaling escan
    scaled_data['escan'] = np.zeros(escan_data.shape)
    scaled_data['escan_eps'] = np.zeros(escan_data_eps.shape)

    ref_tscan_inv_SN = ref_tscan_data_eps/ref_tscan_data 
    ref_tscan_inv_SN = interp1d(t-ref_time_zeros, 
                                ref_tscan_inv_SN, kind='nearest')(escan_time)
    flu_ref = interp1d(t-ref_time_zeros, ref_tscan_data)(escan_time)
    flu_ref = flu_ref/flu_ref[e_ref_index]
    flu_e = interp1d(e, escan_data, axis=0)(ref_tscan_energy)
    flu_e = flu_e/flu_e[e_ref_index]

    #  Scaling factor obtained by tscan have N/S ambiguity
    for i in range(escan_data.shape[1]):
        cond = np.abs(flu_e[i]-flu_ref[i]) > \
               1.0*ref_tscan_inv_SN[i]*np.max([flu_ref[i], flu_e[i]])
        if cond:
            scaled_data['escan'][:, i] = flu_ref[i]/flu_e[i]*escan_data[:, i]
            scaled_data['escan_eps'][:, i] = \
                flu_ref[i]/flu_e[i]*escan_data_eps[:, i]
        else:
            scaled_data['escan'][:, i] = escan_data[:, i]
            scaled_data['escan_eps'][:, i] = escan_data_eps[:, i]

    return scaled_data



'''
tools:
subpackage for TRXASprefitpack utilities

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3
'''

from .TRXASprefitpack_info import TRXASprefitpack_info
from .fit_static import fit_static
from .fit_tscan import fit_tscan
from .auto_scale import auto_scale
from .broadening import broadening

__all__ = ['TRXASprefitpack_info', 'fit_static', 'fit_tscan',
           'auto_scale', 'broadening']



'''
exp_conv_irf:
submodule for fitting data with sum of exponential decay convolved with irf

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3.
'''

import numpy as np
import scipy.linalg as LA
from .A_matrix import make_A_matrix_cauchy
from .A_matrix import make_A_matrix_gau, make_A_matrix_pvoigt


def model_n_comp_conv(t, fwhm, tau, c, base=True, irf='g', eta=None):

    '''
    model for n component fitting
    n exponential function convolved with irf; 'g': normalized gaussian distribution, 'c': normalized cauchy distribution, 'pv': pseudo voigt profile :math:`(1-\\eta)g + \\eta c`

    :param numpy_1d_array t: time
    :param numpy_1d_array fwhm: fwhm of X-ray temporal pulse, if irf == 'g' or 'c' then fwhm = [fwhm], if irf == 'pv' then fwhm = [fwhm_G, fwhm_L]
    :param numpy_1d_array tau: life time for each component
    :param numpy_1d_array c: coefficient
     (num_comp+1,) if base=True
     (num_comp,)   if base=False
    :param base: whether or not include baseline [default: True]
    :type base: bool, optional
    :param irf: shape of instrumental response function [default: g], 
     'g': normalized gaussian distribution, 
     'c': normalized cauchy distribution, 
     'pv': pseudo voigt profile :math:`(1-\\eta)g + \\eta c`

    :type irf: string, optional
    :param eta: mixing parameter for pseudo voigt profile
     (only needed for pseudo voigt profile, Default value is guessed according to Journal of Applied Crystallography. 33 (6): 1311–1316.)
    :type eta: float, optional
    
    :return: fit
    :rtype: numpy_1d_array
    '''

    k = 1/tau
    if base:
        k = np.concatenate((k, np.array([0])))

    if irf == 'g':
        A = make_A_matrix_gau(t, fwhm[0], k)
    elif irf == 'c':
        A = make_A_matrix_cauchy(t, fwhm[0], k)
    elif irf == 'pv':
        if eta is None:
            f = fwhm[0]**5+2.69269*fwhm[0]**4*fwhm[1] + \
                2.42843*fwhm[0]**3*fwhm[1]**2 + \
                4.47163*fwhm[0]**2*fwhm[1]**3 + \
                0.07842*fwhm[0]*fwhm[1]**4 + \
                fwhm[1]**5
            f = f**(1/5)
            x = fwhm[1]/f
            eta = 1.36603*x-0.47719*x**2+0.11116*x**3
        A = make_A_matrix_pvoigt(t, fwhm[0], fwhm[1],
                                 eta, k)

    y = c@A

    return y


def fact_anal_exp_conv(t, fwhm, tau, irf='g', eta=None,
                       data=None, eps=None, base=True):
    
    '''
    Estimate the best coefficiets when fwhm and tau are given

    Before fitting with Escan data, you want fit your model to tscan data
    to know how many component needed for well description of tscan data
    To do this you have good initial guess for not only life time of
    each component but also coefficients.
    For this, It will solve linear least square problem

    :param numpy_1d_array t: time
    :param numpy_1d_array fwhm: fwhm of X-ray temporal pulse, if irf == 'g' or 'c' then fwhm = [fwhm], if irf == 'pv' then fwhm = [fwhm_G, fwhm_L]
    :param numpy_1d_array tau: life time for each component
    :param irf: shape of instrumental response function [default: g], 
     'g': normalized gaussian distribution, 
     'c': normalized cauchy distribution, 
     'pv': pseudo voigt profile :math:`(1-\\eta)g + \\eta c`
    :type irf: string, optional
    :param eta: mixing parameter for pseudo voigt profile 
     (only needed for pseudo voigt profile, Default value is guessed according to Journal of Applied Crystallography. 33 (6): 1311–1316.)
    :type eta: float, optional
    :param numpy_1d_array data: tscan data
    :param numpy_1d_array eps: error for tscan data 
    :param base: whether or not include baseline [default: True]
    :type base: bool, optional
    

    :return: best coefficient for given fwhm, tau
    :rtype: numpy_1d_array
    '''

    k = 1/tau
    if base:
        k = np.concatenate((k, np.array([0])))

    if irf == 'g':
        A = make_A_matrix_gau(t, fwhm[0], k)
    elif irf == 'c':
        A = make_A_matrix_cauchy(t, fwhm[0], k)
    elif irf == 'pv':
        if eta is None:
            f = fwhm[0]**5+2.69269*fwhm[0]**4*fwhm[1] + \
                2.42843*fwhm[0]**3*fwhm[1]**2 + \
                4.47163*fwhm[0]**2*fwhm[1]**3 + \
                0.07842*fwhm[0]*fwhm[1]**4 + \
                fwhm[1]**5
            f = f**(1/5)
            x = fwhm[1]/f
            eta = 1.36603*x-0.47719*x**2+0.11116*x**3
        A = make_A_matrix_pvoigt(t, fwhm[0], fwhm[1], eta, k)

    if eps is not None:
        y = data/eps
        for i in range(k.shape[0]):
            A[i, :] = A[i, :]/eps
    else:
        y = data
    c, _, _, _ = LA.lstsq(A.T, y)

    return c



'''
doc:
Subpackage for the TRXASprefitpack documentation

[deprecated]
This subpackage will be removed in version 0.5

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3.
'''

from .info import __info__

__all__ = ['__info__']



'''
rate_eq:
submodule which solves 1st order rate equation and computes
the solution and signal 

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3.
'''

import scipy.linalg as LA
from .A_matrix import make_A_matrix, make_A_matrix_cauchy
from .A_matrix import make_A_matrix_gau, make_A_matrix_pvoigt


def solve_model(equation, y0):

    '''
    Solve system of first order rate equation

    
    :param equation: matrix corresponding to model
    :type equation: numpy_nd_array
    :param y0: initial condition
    :type y0: numpy_1d_array

    
    :return: eigenvalue of equation
    :rtype: numpy_1d_array
    :return: eigenvectors for equation
    :rtype: numpy_nd_array
    :return: coefficient where y_0 = Vc
    :rtype: numpy_1d_array
    '''

    eigval, V = LA.eig(equation)
    c = LA.solve(V, y0)

    return eigval.real, V, c


def compute_model(t, eigval, V, c):

    '''
    Compute solution of the system of rate equations solved by solve_model
    Note: eigval, V, c should be obtained from solve_model

    
    :param t: time
    :type t: numpy_1d_array
    :param eigval: eigenvalue for equation
    :type eigval: numpy_1d_array
    :param V: eigenvectors for equation
    :type V: numpy_nd_array
    :param c: coefficient
    :type c: numpy_1d_array

    
    :return: solution of rate equation
    :rtype: numpy_nd_array
    '''

    A = make_A_matrix(t, -eigval)
    y = (c * V) @ A
    return y


def compute_signal_gau(t, fwhm, eigval, V, c):

    '''
    Compute solution of the system of rate equations solved by solve_model
    convolved with normalized gaussian distribution
    Note: eigval, V, c should be obtained from solve_model

    
    :param t: time
    :type t: numpy_1d_array
    :param fwhm: full width at half maximum of x-ray temporal pulse
    :type fwhm: float
    :param eigval: eigenvalue for equation
    :type eigval: numpy_1d_array
    :param V: eigenvectors for equation
    :type V: numpy_nd_array
    :param c: coefficient
    :type c: numpy_1d_array
  

    :return: 
     solution of rate equation convolved with normalized gaussian distribution
    :rtype: numpy_nd_array 
    '''

    A = make_A_matrix_gau(t, fwhm, -eigval)
    y_signal = (c * V) @ A
    return y_signal


def compute_signal_cauchy(t, fwhm, eigval, V, c):

    '''
    Compute solution of the system of rate equations solved by solve_model
    convolved with normalized cauchy distribution
    Note: eigval, V, c should be obtained from solve_model

    
    :param t: time
    :type t: numpy_1d_array
    :param fwhm: full width at half maximum of x-ray temporal pulse
    :type fwhm: float
    :param eigval: eigenvalue for equation
    :type eigval: numpy_1d_array
    :param V: eigenvectors for equation
    :type V: numpy_nd_array
    :param c: coefficient
    :type c: numpy_1d_array
  

    :return: 
     solution of rate equation convolved with normalized cachy distribution
    :rtype: numpy_nd_array 
    '''

    A = make_A_matrix_cauchy(t, fwhm, -eigval)
    y_signal = (c * V) @ A
    return y_signal


def compute_signal_pvoigt(t, fwhm_G, fwhm_L, eta, eigval, V, c):

    '''
    Compute solution of the system of rate equations solved by solve_model
    convolved with normalized pseudo voigt profile
    (:math:`pvoigt = (1-\\eta) G(t) + \\eta L(t)`,
    G(t): stands for normalized gaussian
    L(t): stands for normalized cauchy(lorenzian) distribution)

    Note: eigval, V, c should be obtained from solve_model

    :param t: time
    :type t: numpy_1d_array
    :param fwhm_G: 
     full width at half maximum of x-ray temporal pulse (gaussian part)
    :type fwhm_G: float
    :param fwhm_L:
     full width at half maximum of x-ray temporal pulse (lorenzian part)
    :type fwhm_L: float
    :param float eta: mixing parameter :math:`(0 < \\eta < 1)`

    :type eta: float
    :param eigval: eigenvalue for equation
    :type eigval: numpy_1d_array
    :param V: eigenvectors for equation
    :type V: numpy_nd_array
    :param c: coefficient
    :type c: numpy_1d_array

    :return: 
     solution of rate equation convolved with normalized pseudo voigt profile
    :rtype: numpy_nd_array
    '''

    A = make_A_matrix_pvoigt(t, fwhm_G, fwhm_L, eta, -eigval)
    y_signal = (c * V) @ A
    return y_signal

